#include <windows.h>
#include <list>
#include <string>
#include "Snail.h"
#include "SnailClassFactory.h"
#include "SnailShellExt.h"
#include "fileutil.h"
#include "trace.h"

#define INITGUID
#include <initguid.h>
#include <shlguid.h>

// {15236FA4-F792-4ea7-BF7F-B76971F6E11B}
DEFINE_GUID(CLSID_Snail,
0x15236fa4, 0xf792, 0x4ea7, 0xbf, 0x7f, 0xb7, 0x69, 0x71, 0xf6, 0xe1, 0x1b);

HINSTANCE hInst;
std::list<std::string> _include_path;

static void init(void)
{
    std::string path_prefix;

    char *env_path;
    char *env_drive;
    env_path = getenv("HOMEPATH");
    env_drive = getenv("HOMEDRIVE");
    if (env_path && env_drive)
    {
        path_prefix = env_drive;
        path_prefix += env_path;
    }

    if (path_prefix.size())
    {
        if (path_prefix[path_prefix.size() - 1] != euc::PATH_SEPARATOR)
            path_prefix += euc::PATH_SEPARATOR;
        path_prefix += ".kittie";
        path_prefix += euc::PATH_SEPARATOR;
        path_prefix += "include";
        trace("Adding include path: '%s'\n", path_prefix.c_str());
        _include_path.push_back(path_prefix);
    }

    // 获取可执行文件自身路径
    char buffer[1024];
    GetModuleFileName(hInst, buffer, 1024);
    path_prefix = euc::getpathname(buffer);

    path_prefix += "include";
    trace("Adding include path: '%s'\n", path_prefix.c_str());
    _include_path.push_back(path_prefix);
}

extern "C" int APIENTRY DllMain(HINSTANCE hInstance,
    DWORD dwReason, LPVOID)
{
    switch (dwReason)
    {
    case DLL_PROCESS_ATTACH:
        // Extension DLL one-time initialization
        trace("DLL_PROCESS_ATTACH\n");
        hInst = hInstance;
        init();
        break;
    case DLL_PROCESS_DETACH:
        trace("DLL_PROCESS_DETACH\n");
        break;
    }

    return 1;   // OK
}

STDAPI DllCanUnloadNow(void)
{
    return ((SnailClassFactory::CanUnloadNow()
        && SnailShellExt::CanUnloadNow())
        ? S_OK : S_FALSE);
}

STDAPI DllGetClassObject(REFCLSID clsid, REFIID iid, LPVOID *object)
{
    trace("DllGetClassObject()\n");
    *object = NULL;
    SnailClassFactory *pcf = NULL;

    if (IsEqualIID(clsid, CLSID_Snail))
    {
        pcf = new SnailClassFactory();
        return pcf->QueryInterface(iid, object);
    }

    pcf->Release();
    return CLASS_E_CLASSNOTAVAILABLE;
}
