#include "SnailClassFactory.h"
#include "SnailShellExt.h"

ULONG SnailClassFactory::_global_ref_counter = 0L;

SnailClassFactory::SnailClassFactory(void) : _ref_counter(0)
{
    InterlockedIncrement((PLONG) &_global_ref_counter);
}

SnailClassFactory::~SnailClassFactory(void)
{
    InterlockedDecrement((PLONG) &_global_ref_counter);
}

// IUnknown members
STDMETHODIMP SnailClassFactory::QueryInterface(REFIID iid, LPVOID FAR *inter)
{
    *inter = NULL;

    // Any interface on this object is the object pointer

    if (IsEqualIID(iid, IID_IUnknown))
        *inter = (LPUNKNOWN) this;
    else if (IsEqualIID(iid, IID_IClassFactory))
        *inter = (LPCLASSFACTORY) this;

    if (inter)
    {
        AddRef();
        return NOERROR;
    }

    return E_NOINTERFACE;
}

STDMETHODIMP_(ULONG) SnailClassFactory::AddRef(void)
{
    return InterlockedIncrement((LPLONG)& _ref_counter);
}

STDMETHODIMP_(ULONG) SnailClassFactory::Release(void)
{
    if (InterlockedDecrement((LPLONG)& _ref_counter))
        return _ref_counter;
    delete this;
    return 0L;
}

// IClassFactory members
STDMETHODIMP SnailClassFactory::CreateInstance(LPUNKNOWN outter,
    REFIID riid, LPVOID FAR *object)
{
    *object = NULL;

    // Shell extensions typically don't support aggregation (inheritance)
    if (outter)
        return CLASS_E_NOAGGREGATION;

    // Create the main shell extension object.  The shell will then call
    // QueryInterface with IID_IShellExtInit--this is how shell extensions
    // are initialized.
    SnailShellExt *pShellExt = new SnailShellExt();

    if (NULL == pShellExt)
        return E_OUTOFMEMORY;

    HRESULT result = pShellExt->QueryInterface(riid, object);
    if (!object)
        delete pShellExt;

    return result;
}

STDMETHODIMP SnailClassFactory::LockServer(BOOL lock)
{
    return (HRESULT) E_NOTIMPL;
}

bool SnailClassFactory::CanUnloadNow(void)
{
    return _global_ref_counter == 0;
}
