// Last Modified: 2005-08-11 17:38:45
#include <fstream>
#include <vector>
#include <set>
#include "stringutil.h"
#include "fileutil.h"
#include "in_kmk_util.h"
#include "kmk_parser.h"
#include "trace.h"

struct _env
{
public:
    std::string _filename;
    // 0 - 没有定义 platform 环境
    // 1 - 合适的 platform 环境
    // 2 - 不合适的 platform 环境
    int platform_flag;
    // 0 - 初始状态
    // 1 - configuration 状态
    // 2 - file list 状态
    int cur_status;
    std::string _cur_path;
};

std::list<_env> _env_stack;

#define MAX_RECURSIVE_DEPTH 20

bool _read(const char *const filename, std::list<std::string> &_list,
    std::set<std::string> &_set);

bool _parse_pure_line(std::string line, std::list<std::string> &_list,
    std::set<std::string> &_set)
{
    bool success = true;
    std::string filename_buffer;
    // char buffer[4096];

    switch (_env_stack.back().cur_status)
    {
    case 0:
        if (line[0] == '[' && line[line.size() - 1] == ']')
        {
            _env_stack.back().cur_status = 1;
            std::string config_name = line.substr(1, line.size() - 2);
            if ((_set.find(config_name)) == _set.end())
            {
                _set.insert(config_name);
                if (config_name != ".all")
                    _list.push_back(config_name);
            }
            trace("Configuration begin: '%s'\n", config_name.c_str());
        }
        else if (line[line.size() - 1] == ':')
        {
            _env_stack.back().cur_status = 2;
        }
        else
        {
            trace("Error\n");
            return false;
        }
        break;
    case 1:
        if (line[0] == '[' && line[line.size() - 1] == ']')
        {
            _env_stack.back().cur_status = 1;
            std::string config_name = line.substr(1, line.size() - 2);
            if ((_set.find(config_name)) == _set.end())
            {
                _set.insert(config_name);
                if (config_name != ".all")
                    _list.push_back(config_name);
            }
            trace("Configuration begin: '%s'\n", config_name.c_str());
        }
        else if (line[line.size() - 1] == ':')
        {
            _env_stack.back().cur_status = 2;
        }
        break;
    default:
        break;
    }

    return success;
}

bool _parse_line(std::string line, std::list<std::string> &_list,
    std::set<std::string> &_set)
{
    bool success = true;

    // 处理预编译选项，如平台指令等。
    if (line[0] == '#')
    {
        if (line.size() == 1)
        {
            trace("Error\n");
            return false;
        }

        // 将预编译选项分离为若干 tokens

        std::vector<std::string> params;

        // 0 - token 之前
        // 1 - token 之中
        bool in_command = false;
        // 双引号或尖括号指示文件名，不支持转义。
        bool quoted = false;
        std::string command;
        char quoted_char = '\"';
        for (std::string::size_type pos = 1; pos != line.size(); ++pos)
        {
            if (quoted)
            {
                if (line[pos] != quoted_char)
                    command += line[pos];
                else
                {
                    quoted = false;
                    params.push_back(command);
                    command = std::string();
                }
            }
            else
            {
                if (in_command)
                {
                    switch (line[pos])
                    {
                    case '\t':
                    case ' ':
                        in_command = false;
                        if (command.size())
                        {
                            params.push_back(command);
                            command = std::string();
                        }
                        break;
                    case '<':
                    case '\"':
                        if (line[pos] == '<')
                            quoted_char = '>';
                        else
                            quoted_char = '\"';
                        in_command = true;
                        quoted = true;
                        if (command.size())
                        {
                            params.push_back(command);
                            command = std::string();
                        }
                        break;
                    default:
                        command += line[pos];
                        break;
                    }
                }
                else
                {
                    switch (line[pos])
                    {
                    case '\t':
                    case ' ':
                        break;
                    case '<':
                    case '\"':
                        if (line[pos] == '<')
                            quoted_char = '>';
                        else
                            quoted_char = '\"';
                        in_command = true;
                        quoted = true;
                        break;
                    default:
                        command += line[pos];
                        in_command = true;
                        break;
                    }
                }
            }
        }

        if (command.size())
        {
            params.push_back(command);
            command = std::string();
        }

        // 处理平台指令

        switch (_env_stack.back().platform_flag)
        {
        case 0:
            if (params[0] == "include")
            {
                trace("quoted_char = '%c'\n", quoted_char);
                if (params.size() > 1)
                {
                    if (_env_stack.size() <= MAX_RECURSIVE_DEPTH)
                    {
                        std::string filename;
                        if (quoted_char == '\"')
                        {
                            filename = path_simplify(
                                getpathname(_env_stack.back()._filename),
                                params[1], "win32");
                            success = _read(filename.c_str(), _list, _set);
                        }
                        else
                        {
                            std::list<std::string>::iterator ite;
                            for (ite = _include_path.begin();
                                ite != _include_path.end(); ++ite)
                            {
                                filename = *ite;
                                filename += euc::PATH_SEPARATOR;
                                filename += params[1];

                                trace("Checking for '%s'\n",
                                    filename.c_str());
                                if (euc::file_exists(filename) == 2)
                                {
                                    success = _read(filename.c_str(), _list, _set);
                                    break;
                                }
                                trace("Not exists...\n");
                            }
                        }
                    }
                    else
                    {
                        success = false;
                    }
                }
                return success;
            }

            if (params[0] != "platform")
            {
                trace("Error\n");
                return false;
            }

            _env_stack.back().platform_flag = 2;
            for (std::vector<std::string>::size_type i = 1;
                i < params.size(); ++i)
                if (params[i] == "win32")
                    _env_stack.back().platform_flag = 1;

            break;
        case 1:
        case 2:
            if (params[0] != "end" || params.size() < 2
                || params[1] != "platform")
            {
                trace("Error\n");
                return false;
            }

            _env_stack.back().platform_flag = 0;
            break;
        default:
            break;
        }

        return success;
    }

    if (_env_stack.back().platform_flag == 2)
        return true;

    return _parse_pure_line(line, _list, _set);
}

bool work(std::istream &input, std::list<std::string> &_list,
    std::set<std::string> &_set)
{
    char buffer[4096];

    bool success = true;
    _env_stack.back().platform_flag = 0;
    _env_stack.back().cur_status = 0;
    _env_stack.back()._cur_path = "/";

    // 将配置文件读入内存并分析
    while (input.good())
    {
        // 读取行，清除白空格，记录行号
        input.getline(buffer, 4096);
        std::string line(buffer);
#if defined(LINUX) || defined(UNIX)
        if (line[line.size() - 1] == '\r')
        {
            line.resize(line.size() - 1);
            while (line[line.size() - 1] == ' '
                || line[line.size() - 1] == '\t')
                line.resize(line.size() - 1);
        }
#endif // LINUX || UNIX

        // 排除注释行
        if (line[0] == ';' || line.size() == 0)
            continue;

        // 处理行
        success &= _parse_line(line, _list, _set);
    }

    return success;
}

bool _read(const char *const filename, std::list<std::string> &_list,
    std::set<std::string> &_set)
{
    trace("Open input file '%s'\n", filename);
    std::fstream *input = NULL;
    bool result = false;
    _env env;
    env._filename = euc::replace_all(filename, '\\', '/');
    trace("env._filename = '%s'\n", env._filename.c_str());
    _env_stack.push_back(env);

    try
    {
        input = new std::fstream(filename, std::ios::in);
        if (!input->good())
        {
            input->close();
            delete input;
            _env_stack.pop_back();
            return false;
        }
        result = work(*input, _list, _set);
    }
    catch (...)
    {
        input->close();
        delete input;
        _env_stack.pop_back();
        return false;
    }

    input->close();
    delete input;
    trace("Parse %s\n", result ? "OK" : "error");
    _env_stack.pop_back();
    return result;
}

// Get all configuration in specified kmk file
std::list<std::string> parse_kmk(std::string filename)
{
    std::list<std::string> result;
    std::set<std::string> purify;

    if (_read(filename.c_str(), result, purify))
        return result;

    return std::list<std::string>();
}
